. .\0_globals.ps1

Write-host "Starting the app migration process" -ForegroundColor DarkGray

qlik context use $QlikSaaSContext | out-null

$tenanturl = qlik context get | Select-String -Pattern "server:"
$tenanturl = $tenanturl.ToString().replace('server: ','')
$cliuserid = qlik user me -q

## Define the names of the temporary spaces here
# General temporary space that will be the target of all new app imports in the tenant
$migrationtempspacename = 'MigrationTemp'
$removemigrationtemp = $true
# Shared Space that will host the apps published to a Managed Space so that you can review them and relocate them to their corresponding shared spaces
$publishedappsaharedplaceholder = 'Migration Tools - Published Apps Temporary Placeholder'


$appsdir = "$DATA\Apps"


if($MIGRATIONFROM -eq "QlikSense")
{
	#Qlik Sense - StreamName,AppName,AppID,AppOwnerEmail
	$apps = Import-Csv -Path $DATA\apps.csv -Header "StreamName", "AppName", "AppID", "AppOwnerEmail" | Select-Object -skip 1
}
elseif($MIGRATIONFROM -eq "QlikView")
{
	#QSAppName,AppOwner1,QVAppId1,SpaceName,SpaceId1
    $apps = Import-Csv -Path $DATA\apps.csv -Header  "AppName", "AppOwnerSubject", "AppID", "StreamName", "SpaceID"  | Select-Object -skip 1
}
else 
{
    Write-Host "Invalid MIGRATIONFROM Value." -ForegroundColor Red
    exit;
}

# Function that evaluates if the qlik-cli user can upload contents to a designated space. 
# If the user cannot all roles will be assigned. This is made so that the same user can go directly to the tenant to verify the uploaded apps using the GUI without restrictions.
# Note: As a best practice, once the migration process is concluded, please make sure to remove unecessary roles give for the qlik-cli user to the spaces
# $evaluatedpacename -> name of the space to verify
# $evaluatedspaceuserid -> tenant userid to which the verification should be made to
# $createmissing -> flag to identify if the space must be created in case it is missing in the tenant
# $evaluatedspacecreatetype -> type of space to create in case $createmissing flag is true (shared or managed)
# $evaluatedspacedescription -> description to include when creating a new space
# $evaluatedspacecmdinfotext -> message to output during this evaluation, when required
function cliuserspaceuploadprivileges(){
	Param([string] $evaluatedpacename, [string] $evaluatedspaceuserid, [bool] $createmissing, [string] $evaluatedspacecreatetype, [string] $evaluatedspacedescription, [string] $evaluatedspacecmdinfotext)

	$res = @{}
	$res.Add("tmpspacehousekeeping", $true)
	$res.Add("id",$null)
	$res.Add("spacetype",$null)
	$res.Add("spaceownerid",$null)

	$friendlyusertext = ""
	if($evaluatedspaceuserid -eq $cliuserid){
		$friendlyusertext = "qlik-cli user"
	}else{
		$friendlyusertext = "expected app owner"
	}

	# Internal function that adds a new user (assignee) to a space
	function createroles(){
		Write-host "Info: The $($friendlyusertext) does not have privileges to upload/edit contents to the space '$($evaluatedpacename)'. Adding the necessary roles. Once you have reviewed the uploaded contents pelase make sure to remove unnecessary roles given for the $($friendlyusertext) to the space." -ForegroundColor Cyan
		if($evaluatedspace.type -eq "shared"){
			$migrationtemproles = qlik space assignment create --spaceId $evaluatedspace.id --assigneeId $evaluatedspaceuserid --type user --roles "codeveloper,consumer,dataconsumer,facilitator,producer" | ConvertFrom-Json
		}else{
			$migrationtemproles = qlik space assignment create --spaceId $evaluatedspace.id --assigneeId $evaluatedspaceuserid --type user --roles "basicconsumer,consumer,contributor,dataconsumer,facilitator,publisher" | ConvertFrom-Json
		}
	}

	# Internal function that updates the user (assignee) roles to a space
	function updateroles(){
		Write-host "Info: The $($friendlyusertext) does not have privileges to upload/edit contents to the space '$($evaluatedpacename)'. Adding the necessary roles. Once you have reviewed the uploaded contents pelase make sure to remove unnecessary roles given for the $($friendlyusertext) to the space." -ForegroundColor Cyan
		if($evaluatedspace.type -eq "shared"){
			$migrationtemproles = qlik space assignment update $evaluatedspaceuseridfounddetails.id --spaceId $evaluatedspace.id --roles "codeveloper,consumer,dataconsumer,facilitator,producer" | ConvertFrom-Json
		}else{
			$migrationtemproles = qlik space assignment update $evaluatedspaceuseridfounddetails.id --spaceId $evaluatedspace.id --roles "basicconsumer,consumer,contributor,dataconsumer,facilitator,publisher" | ConvertFrom-Json
		}
	}

	# $evaluatedspace = qlik space ls --name $evaluatedpacename | ConvertFrom-Json
	$evaluatedspace = qlik space ls --filter ('name eq \"' + $evaluatedpacename + '\"') | ConvertFrom-Json

	if($null -eq $evaluatedspace.id){
		if($createmissing){
			Write-host "Info: The space '$($evaluatedpacename)' does not exist in the tenant. Creating it. $($evaluatedspacecmdinfotext)" -ForegroundColor DarkGray
			$evaluatedspace = qlik space create --name $evaluatedpacename --type $evaluatedspacecreatetype --description $evaluatedspacedescription | ConvertFrom-Json
		}else{
			# Write-host $evaluatedspacecmdinfotext -ForegroundColor Magenta
			$res.tmpspacehousekeeping = $false
			return $res
		}
	}elseif($evaluatedspace.ownerId -ne $evaluatedspaceuserid){
		# Seems the space already exists and the owner is not the same user as the qlik-cli user. Adding roles to allow app upload
		$migrationtemproles =  qlik space assignment ls --spaceId $evaluatedspace.id | ConvertFrom-Json

		if($migrationtemproles.length -eq 0){
			createroles
		}else{
			$evaluatedspaceuseridfound = $false
			$evaluatedspaceuseridfounddetails = @{}

			foreach($migrationtempuser in $migrationtemproles ){
				if($evaluatedspaceuserid -eq $migrationtempuser.assigneeId){
					$evaluatedspaceuseridfound = $true
					$evaluatedspaceuseridfounddetails = $migrationtempuser
				}
			}

			if($evaluatedspaceuseridfound){
				if($null -eq $evaluatedspaceuseridfounddetails.roles){
					createroles
				}else{
					$migrationtempuploadprivileges = $false
	
					foreach($role in $evaluatedspaceuseridfounddetails.roles){
						if($role -eq "codeveloper" -or 	$role -eq "facilitator" -or $role -eq "producer" -or $role -eq "publisher"){
							$migrationtempuploadprivileges = $true
						}
					}

					if($migrationtempuploadprivileges -ne $true){
						updateroles
					}
				}
			}else{
				createroles
			}
		}
	}

	$res.id = $evaluatedspace.id
	$res.spacetype = $evaluatedspace.type
	$res.spaceownerid = $evaluatedspace.ownerid

	return $res
}

# Ensuring temporary space is available and ready to use
$migrationspace = cliuserspaceuploadprivileges $migrationtempspacename $cliuserid $true "shared" "This is a temporary space created by the 'Migration Tools'. Unless something went wrong, it should be automatically deleted once the 7_migrateapps.ps1 script finishes its execution." "This is a temporary space to support the app migration process. If all app uploads is successful it will be removed at the end."
$migrationspaceid = $migrationspace.id
$removemigrationtemp = $migrationspace.tmpspacehousekeeping

# Based on code from
# https://qlik.dev/tutorials/migrate-apps-from-qlik-sense-on-windows-to-qlik-sense-saas#sample-code
# https://qlik.dev/tutorials/migrate-community-sheets-from-qlik-sense-windows-to-saas
# https://qlik.dev/manage/migrate/qlik-cli-app-objects-migrate/

foreach ($app in $apps)
{
	$spacename = $app.StreamName
	$appname = $app.AppName
	$appid = $app.AppID

	Write-Host
	Write-Host "Migrating App '$($appname)' with ID '$($appid)'" -ForegroundColor Green

	if($MIGRATIONFROM -eq "QlikSense")
	{
		$appowneremail = $app.AppOwnerEmail
		$userid =  qlik user ls --email $appowneremail -q
	}
	elseif($MIGRATIONFROM -eq "QlikView")
	{
		if($app.AppOwnerSubject -ne "-"){
			$subject = $app.AppOwnerSubject
			$userid =  qlik user ls --subject $subject -q	
		}else{
			Write-Host "Info: there is no app owner defined for this app. Assuming the qlik-cli user as owner." -ForegroundColor DarkGray
			$userid = $cliuserid
		}
	}

	if($null -ne $userid){
		Write-Host "Info: the app owner tenant User ID is '$($userid)'" -ForegroundColor DarkGray
		
		$exportAppFile = ""
		$appobjects = @{}
		# Exporting QVFs
		if($MIGRATIONFROM -eq "QlikSense"){
			$exportAppFile = "$($appsdir)\$($appid).qvf"
			qlik context use $QlikWindowsContext | out-null
			qlik qrs app export create $appid --skipdata $SS --exportScope all --output-file $exportAppFile
			$appobjects = qlik qrs app object full --filter "app.id eq $appid" --insecure | ConvertFrom-Json
			qlik context use $QlikSaaSContext | out-null
		}
		elseif($MIGRATIONFROM -eq "QlikView"){
			# you should be placing the apps you convert using the qlikview converter tool into $appsdir (i.e. Data\Apps)
			# rvr: review this warning, reuse the $exportAppFile logic. 
			$exportAppFile = "$($appsdir)\$($appname).qvf"
		}
		
		# Identifying the target space characteristics
		$spacetype = ""
		$spaceid = ""
		$spaceownerid = ""
		$spacefriendlyname = ""
		$spaceexists = $true
		if($spacename -eq '-' -or $spacename -eq 'Work')
		{
			$spacetype = "personal"
			$spacefriendlyname = "the owner's 'Personal' space"
		}else{
			$targetspaceid = cliuserspaceuploadprivileges $spacename $cliuserid $false "" "" ""

			if($null -eq $targetspaceid.spaceownerid){
				$spaceexists = $false
			}

			$spacetype = $targetspaceid.spacetype
			$spaceid = $targetspaceid.id
			$spaceownerid = $targetspaceid.spaceownerid
			$spacefriendlyname = "the space '$($spacename)'"
		}

		$cloudappid = ""
		if($spaceexists){
			$cloudapp = qlik app import -f $exportAppFile --spaceId $migrationspaceid | ConvertFrom-Json
			$cloudappid = $cloudapp.attributes.id
		}

		$appobjectsassignations = @()
		if($MIGRATIONFROM -eq "QlikSense" -and $spaceexists){
			if($null -ne $appobjects){
				Write-Host "Looking for contents on the app that will need to be under ownership revision/assignment." 
				Write-Host "Currently looking for Bookmarks, Sheets and Stories. Depending on the app complexity this might take a bit." -ForegroundColor DarkGray
				foreach($i in $appobjects){
					if($i.objectType -eq "sheet" -or $i.objectType -eq "story" -or $i.objectType -eq "bookmark"){
						$appobjectownership = @{}
	
						# Retrieve the object and owner
						# If your user subjects in Qlik Cloud don't match directory\userId, use a different approach
						$appobjectownership.Add("onpremid", $i.engineObjectId)
						$appobjectownership.Add("onpremsubject", $i.owner.userDirectory + '\\' + $i.owner.userId)
						$appobjectownership.Add("onpremdescription", $i.description)
						$cloudUser = qlik user ls --filter ('subject eq \"' + $appobjectownership.onpremsubject + '\"') | ConvertFrom-Json
						$appobjectownership.Add("cloudtargetuserid", $cloudUser.id)
						
						# Additional data that will be useful later
						$appobjectownership.Add("objecttype", $i.objectType)
						$appobjectlayout = @{}
						if($i.objectType -eq "sheet" -or $i.objectType -eq "story"){
							$appobjectlayout = qlik app object layout $appobjectownership.onpremid -a $cloudappid | ConvertFrom-Json
						}elseif($i.objectType -eq "bookmark"){
							$appobjectlayout = qlik app bookmark layout $appobjectownership.onpremid -a $cloudappid | ConvertFrom-Json
						}
						$appobjectownership.Add("cloudlayoutownerid", $appobjectlayout.qMeta.ownerId)
						$appobjectownership.Add("cloudlayoutispublished", $appobjectlayout.qMeta.published)
						$appobjectownership.Add("cloudlayoutisapproved", $appobjectlayout.qMeta.approved)
						$appobjectownership.Add("objectflaggedasprivate", $appobjectownership.cloudlayoutispublished -ne $true)
						$appobjectownership.Add("objectflaggedasownerupdateneeded", $appobjectownership.cloudtargetuserid -ne $appobjectownership.cloudlayoutownerid -and $appobjectownership.cloudlayoutispublished -ne $true)
	
						$appobjectsassignations = $appobjectsassignations + $appobjectownership
					}
				} 
	
				Write-Host "Found $($appobjectsassignations.Count) app contents that will be under ownership revision/assignment."
			}
		}

		$assignedspace = ""
		if($spacetype -eq "personal"){
			if($MIGRATIONFROM -eq "QlikSense"){
				Write-Host "Reviewing ownership assignments for Bookmarks, Sheets and Stories found."
				Write-Host "Depending on the app complexity this might take a bit." -ForegroundColor DarkGray
				foreach($appobject in $appobjectsassignations){	
					# Ensuring correct assignation of object/bookmark ownership
					if($appobject.objectflaggedasownerupdateneeded -eq $true){
						qlik app object change-owner --appId $cloudappid --objectId $appobject.onpremid --ownerId $userid | out-null
					}
				}
			}
			
			Write-Host "App is targeted to be imported into the user's 'personal' space"
			$assignedspace = qlik app space rm $cloudappid | ConvertFrom-Json

			# Assigning the rightful app ownership
			$assignedownership = qlik app owner $cloudappid --ownerId $userid | ConvertFrom-Json

		}elseif(($spacetype -eq "shared" -or $spacetype -eq "managed") -and $spaceid -ne "" -and $null -ne $spaceid -and $spaceexists){
			Write-Host "App is targeted to be imported into a '$($spacetype)' space"

			if($spacetype -eq "managed"){
				# Duplicating the app, which will end up to become the source app for the published one in the managed space
				$appcopyid = (qlik app copy $cloudappid | ConvertFrom-Json).attributes.id
			}
			
			if($MIGRATIONFROM -eq "QlikSense"){
				Write-Host "Reviewing ownership assignments for Bookmarks, Sheets and Stories found."
				Write-Host "Depending on the app complexity this might take a bit." -ForegroundColor DarkGray
				foreach($appobject in $appobjectsassignations){	
					# Ensuring correct assignation of object/bookmark ownership
					if($appobject.objectflaggedasownerupdateneeded -eq $true){
						qlik app object change-owner --appId $cloudappid --objectId $appobject.onpremid --ownerId $appobject.cloudtargetuserid  | out-null
					}
				}
			}

			if($spacetype -eq "shared"){
				$assignedspace = qlik app space update $cloudappid --spaceId $spaceid | ConvertFrom-Json
			}else{
				# Placing the app(s) in the temporary shared space. This space will persist after running the script so you can move the source app to a space of your choosing
				Write-Host "Creating the source app in the supporting shared space named '$($publishedappsaharedplaceholder)'."
				Write-Host "It is recommended that once the migration is completed, you review the source app and move it to another shared space." -ForegroundColor DarkGray

				$managedplaceholderspace = cliuserspaceuploadprivileges $publishedappsaharedplaceholder $cliuserid $true "shared" "Space created by the 'Migration Tools' to host the source applications that have been targeted to a Managed Space. Please move the apps from this space to their target shared space and delete the space once all these source apps are on their inteded destinations. The relationship between these source apps and their published counterparts will be kept if moved to a new shared space." ""
							
				qlik app space update $cloudappid --spaceId $managedplaceholderspace.id | out-null
				$copyassignedspace = qlik app space update $appcopyid --spaceId $managedplaceholderspace.id | ConvertFrom-Json

				# Publishing the app to the target managed space
				$assignedspace = qlik app publish create $cloudappid --moveApp --originAppId $appcopyid --spaceId $spaceid | ConvertFrom-Json

				# Removing personal objects from the source app (in the shared space)
				if($MIGRATIONFROM -eq "QlikSense"){
					Write-Host "Performing a quick housekeeping to the source app by removing all 'non-approved' contents: "
					foreach($appobject in $appobjectsassignations){	
						if($appobject.cloudlayoutisapproved -eq $false){
							if($appobject.objecttype -ne "bookmark"){
								if($appobject.cloudlayoutispublished -eq $true){
									qlik app object unpublish $appobject.onpremid --app $appcopyid | out-null
								}
								qlik app object rm $appobject.onpremid --app $appcopyid | out-null
							}else{
								if($appobject.cloudlayoutispublished -eq $true){
									qlik app bookmark unpublish $appobject.onpremid --app $appcopyid | out-null
								}
								qlik app bookmark rm $appobject.onpremid --app $appcopyid | out-null
							}
						}
					}
				}

				# Ensuring the app owner also has access to the temporary shared space and is owner of the source app
				$targetspaceid = cliuserspaceuploadprivileges $publishedappsaharedplaceholder $userid $false "" "" ""
				$assignedownership = qlik app owner $appcopyid --ownerId $userid | ConvertFrom-Json
			}

			# Assigning the rightful app ownership
			$targetspaceid = cliuserspaceuploadprivileges $spacename $userid $false "" "" ""
			$assignedownership = qlik app owner $cloudappid --ownerId $userid | ConvertFrom-Json
		
		}else{
			Write-Host "Warning: the app is targeted to be imported in a space that does not exist in the tenant. Please add the '$($spacename)' space to your tenant and re-import the app." -ForegroundColor Magenta
		}

		# Some Housekeeping
		if($MIGRATIONFROM -eq "QlikSense"){
			Remove-Item $exportAppFile # rvr: maybe make this Qlik Sense only
		}else{
			Write-Host "Info: QlikView Migrated apps need to be removed manually from their folder location [$($exportAppFile)]" -ForegroundColor DarkGray
		}
		
		if($spaceexists){
			Write-Host "Done: the app '$($appname)' is now migrated to $($spacefriendlyname)" -ForegroundColor Green
			
			if($spacetype -eq "personal"){
				Write-Host "App url (for the app owner only) [$($tenanturl)/sense/app/$($cloudappid)]" -ForegroundColor DarkGray
				Write-Host "Space url (for the app owner only) [$($tenanturl)/catalog?space_filter=personal]" -ForegroundColor DarkGray
			}
			else{
				Write-Host "App url [$($tenanturl)/sense/app/$($cloudappid)]" -ForegroundColor DarkGray
				Write-Host "Space url [$($tenanturl)/catalog?space_filter=$($spaceid)]" -ForegroundColor DarkGray
			}
			
			if($spacetype -eq "managed"){
				Write-Host "Source app url [$($tenanturl)/sense/app/$($appcopyid)]" -ForegroundColor DarkGray
				Write-Host "Source app (temporary) space url [$($tenanturl)/catalog?space_filter=$($managedplaceholderspace.id)]" -ForegroundColor DarkGray
			}
		}
	}else{
		Write-Host "Warning: Unable to retrieve user ID in the tenant for user email '$($appowneremail)'. Skipping this app." -ForegroundColor Magenta
	}
}

# Some Housekeeping
Write-Host
if($removemigrationtemp){
	Write-Host "Info: the space $($migrationtempspacename) is no longer needed. Removing it. " -ForegroundColor DarkGray
	qlik space rm $migrationspaceid
}else{
	Write-Host "Warning: the space $($migrationtempspacename) will persist on your tenant since contains apps that could not be moved to their space destination. Please review them and remove the space manually. " -ForegroundColor Magenta
}